import java.util.Timer
import java.util.TimerTask
import scala.util.control.NonFatal
import scala.util.Try
import scala.concurrent.Future
import scala.concurrent.ExecutionContext.Implicits.global
import scala.collection.mutable

sealed trait Observable[T]:
  def subscribe(observer: Observer[T]): Subscription
    
  def foreach(f: T => Unit) = subscribe(Observer(f))
  
  def take(count: Int) = Observable[T] { obs =>
    var n = 1
    val observer= Observer[T](
      t => if (n < count) { n += 1; obs.onNext(t) } else if (n == count) {
        obs.onNext(t); obs.onComplete()
      },
      obs.onError,
      obs.onComplete()
    ) 
    subscribe(observer)
  }
  
  def repeat: Observable[T] = Observable[T] { obs =>
    val subscription = MultipleAssignmentSubscription()
    lazy val observer: Observer[T] = Observer[T](obs.onNext, obs.onError, 
        subscription.subscription = this.subscribe(observer))
    subscription.subscription = subscribe(observer)
    subscription
  }
  
  def onErrorResumeNext(other: => Observable[T]) = Observable[T] { obs =>
    val subscription = MultipleAssignmentSubscription()
    lazy val observer: Observer[T] = Observer[T](obs.onNext, _ => subscription.subscription = other.subscribe(obs), obs.onComplete())
    subscription.subscription = subscribe(observer)
    subscription
  }

  def onErrorReturn(value: T) = Observable[T] { obs =>
    lazy val observer: Observer[T] = Observer(
    obs.onNext, 
    _ => 
      obs.onNext(value)
      obs.onComplete(),
    obs.onComplete())
    this.subscribe(observer)
  }

  def map[P](f: T => P) = Observable[P] { obs =>
    val observer: Observer[T] = Observer[T](
      t => Try(f(t)).fold(obs.onError,obs.onNext),
      obs.onError,
      obs.onComplete())
    subscribe(observer)
  }

  def last: Observable[T] = Observable[T] { obs =>
    var lastSeen: Option[T] = None
    val observer: Observer[T] = Observer[T](
      t => lastSeen = Some(t),
      obs.onError,lastSeen.fold(obs.onComplete()){x => obs.onNext(x); obs.onComplete() }
    )
    subscribe(observer)
  }

  def scan[U](z: U)(f: (T,U) => U) = Observable[U] { obs =>
    var state = z
    obs.onNext(state)
    val observer = Observer[T](
      t =>
        try {
          state = f(t,state)
          obs.onNext(state)
        } catch {
          case e => obs.onError(e)
        },
      obs.onError,
      obs.onComplete()
    )
    this.subscribe(observer)
  }

  def zip[U](that: Observable[U]) = Observable[(T,U)] { obs =>
    val tBuffer = mutable.Queue.empty[T]
    val uBuffer = mutable.Queue.empty[U]
    val tObserver = Observer[T](
      t => 
        (tBuffer,uBuffer) synchronized {
          if uBuffer.nonEmpty then obs.onNext((t,uBuffer.dequeue))
          else tBuffer.enqueue(t)
        },
      obs.onError,
      obs.onComplete()
    )
    val uObserver = Observer[U](
      u => 
        (tBuffer,uBuffer) synchronized {
          if tBuffer.nonEmpty then obs.onNext((tBuffer.dequeue,u))
          else uBuffer.enqueue(u)
        },
      obs.onError,
      obs.onComplete()
    )
    val tSub = this.subscribe(tObserver)
    val uSub = that.subscribe(uObserver)
    Subscription {
      tSub.unsubscribe()
      uSub.unsubscribe()
    }
  }

object Observable:
  def apply[T](create: Observer[T] => Subscription) = new Observable[T]:
    def subscribe(obs: Observer[T]): Subscription =
      var unsubscribed = false
      lazy val observer: Observer[T] = Observer[T](
          t => if (!unsubscribed) obs.onNext(t),
          e => if (!unsubscribed) {
            subscription.unsubscribe()
            obs.onError(e)
          }, if (!unsubscribed) {
            subscription.unsubscribe()
            obs.onComplete()
          })
      lazy val underlying = create(observer)
      lazy val subscription = Subscription {        
        unsubscribed = true
        underlying.unsubscribe()
      }
      underlying; subscription    
  
  def from[T](seq: Iterable[T])(implicit scheduler: Scheduler) = Observable[T] { obs =>
    scheduler.schedule {
      seq.foreach(obs.onNext)
      obs.onComplete()
    }
  }
  
  def interval(delay: Long) = Observable[Long] { obs =>
    val timer = new Timer
    var counter = 0l
    lazy val task: TimerTask = new TimerTask:
      def run(): Unit =
        try { obs.onNext(counter) }
        catch {
          case NonFatal(e) => s.unsubscribe(); obs.onError(e)
        }
        counter += 1          
    timer.schedule(task, delay, delay)
    lazy val s = Subscription { timer.cancel() }
    s
  }
